package main

import (
	`math/big`
	`net`
	"net/netip"
)

// subnetResult is only used for human/"pretty" printing.
type subnetResult netip.Prefix

type tableOpts struct {
	Plain    bool `short:"p" long:"plain" description:"Show plain table output."`
	Legacy   bool `short:"l" long:"legacy" description:"Include legacy/obsolete/deprecated information."`
	NoV4Mask bool `short:"M" long:"no-mask" description:"Do not include netmasks for IPv4."`
	NoIpv6   bool `short:"4" long:"ipv4" description:"Only show IPv4 table(s)."`
	NoIpv4   bool `short:"6" long:"ipv6" description:"Only show IPv6 table(s)."`
}

type tableAddrRet struct {
	Sizer *tableAddrSizer
	Rows  []tableAddr
}

type tableAddr struct {
	Prefix    uint8
	Bits      uint8
	Addresses *big.Int
	Hosts     *big.Int
	NetPrefix netip.Prefix `render:"-"`
}

// tableAddrSizer is used to control spacing/sizing of a tableAddr table's columns.
type tableAddrSizer struct {
	Prefix    uint8
	Bits      uint8
	Addresses uint8
	Hosts     uint8
}

type tableMask4Ret struct {
	Sizer *tableMask4Sizer
	Rows  []tableMask4
}

// tableMask4 is used to hold string representation of netmask information.
type tableMask4 struct {
	Prefix  uint8
	Netmask string
	Hex     string
	Dec     uint32
	Bin     string
	Mask    net.IPMask `render:"-"`
}

// tableMask4Sizer, like tableAddrSizer, is used to control spacing/sizing of a tableMask4 table's columns.
type tableMask4Sizer struct {
	Prefix  uint8
	Netmask uint8
	Hex     uint8
	Dec     uint8
	Bin     uint8
}

type tableLegacy4Ret struct {
	Sizer *tableLegacy4Sizer
	Rows  []tableLegacy4
}

// tableLegacy4 contains a spec for a class in the legacy "classed" IPv4 networking.
type tableLegacy4 struct {
	Class    string
	CIDR     string
	Start    string
	End      string
	NetStart netip.Addr   `render:"-"`
	NetEnd   netip.Addr   `render:"-"`
	NetCIDR  netip.Prefix `render:"-"`
}

// tableLegacy4Sizer is used to size tableLegacy4 entries.
type tableLegacy4Sizer struct {
	Class uint8
	CIDR  uint8
	Start uint8
	End   uint8
}

// tableFormatter is used for "rendering" table output.
type tableFormatter struct {
	// Headers...
	// First char, first line
	TopLeftHdr string
	// Char to fill between TopLeftHdr and TopColSepHdr.
	TopFillHdr string
	// Column separator, first line
	TopColSepHdr string
	// Last char, first line
	TopRightHdr string
	// Column separator for both column separators (title line) and left/right-most lines.
	ColSepHdr string
	// First char, last line of header
	BottomLeftHdr string
	// Char to fill between BottomLeftHdr and BottomColSepHdr.
	BottomFillHdr string
	// Column separator, last line of header
	BottomColSepHdr string
	// Last char, last line of header
	BottomRightHdr string
	// Rows...
	// Left-most line (border).
	Left string
	// Fill the cell lines for values
	Fill string
	// Separate line/cell border columns
	LineColSep string
	// "In-between" rows; left-most.
	LineLeft string
	// "In-between" rows, right-most.
	LineRight string
	// Separate value columns
	ColSep string
	// Right-most line (border)
	Right string
	// Last lines get special treatment as they are also a border.
	LastLeft  string
	LastFill  string
	LastSep   string
	LastRight string
	// This is mostly for experimentation for Plain output, but...
	SuppressLineSep bool
	NoUpperTitle    bool
	NoBoldTitle     bool
}
